/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 * Emulate asprintf().
 *
 ******************************************************************************/

#include <stdio.h>

static int
asprintf(char **r_result, const char *a_format, ...)
{
    int retval;
    char *result;
    va_list ap;
#define CW_ASPRINTF_STARTLEN 16

    result = (char *) malloc(CW_ASPRINTF_STARTLEN);
    if (result == NULL)
    {
	retval = -1;
	goto RETURN;
    }

    va_start(ap, a_format);
    retval = vsnprintf(result, CW_ASPRINTF_STARTLEN, a_format, ap);
    va_end(ap);

    if (retval == -1)
    {
	goto RETURN;
    }
    else if (retval >= CW_ASPRINTF_STARTLEN)
    {
	free(result);
	result = (char *) malloc(retval + 1);
	if (result == NULL)
	{
	    retval = -1;
	    goto RETURN;
	}

	va_start(ap, a_format);
	retval = vsnprintf(result, retval + 1, a_format, ap);
	va_end(ap);
    }

    *r_result = result;
    RETURN:
    return retval;
}
