/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#include "../include/libonyx/libonyx.h"

void
envdict_l_populate(cw_nxo_t *a_dict, cw_nxo_t *a_tkey, cw_nxo_t *a_tval,
		   char **a_envp)
{
    int i;
    char *val_str, *t_str;
    uint32_t key_len, val_len;

    nxo_dict_new(a_dict, true, CW_LIBONYX_ENVDICT_SIZE);

    if (a_envp != NULL)
    {
	/*
	 * Iterate through key/value strings in the environment vector
	 * and insert them into the dictionary.
	 */
	for (i = 0; a_envp[i] != NULL; i++)
	{
	    /* Find the '=' that separates key and value. */
	    val_str = strchr(a_envp[i], '=');
	    key_len = val_str - a_envp[i];
	    val_str++;

	    /* Create key. */
	    nxo_name_new(a_tkey, a_envp[i], key_len, false);

	    /* Create value. */
	    val_len = strlen(val_str);
	    nxo_string_new(a_tval, true, val_len);
	    t_str = nxo_string_get(a_tval);
#ifdef CW_THREADS
	    nxo_string_lock(a_tval);
#endif
	    memcpy(t_str, val_str, val_len);
#ifdef CW_THREADS
	    nxo_string_unlock(a_tval);
#endif

	    /* Insert into dictionary. */
	    nxo_dict_def(a_dict, a_tkey, a_tval);
	}
    }
}
