/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#define CW_NXO_CLASS_C_

#include "../include/libonyx/libonyx.h"
#include "../include/libonyx/nxa_l.h"
#include "../include/libonyx/nxo_l.h"
#include "../include/libonyx/nxo_class_l.h"

void
nxo_class_new(cw_nxo_t *a_nxo, void *a_opaque,
	      cw_nxo_class_ref_iter_t *a_ref_iter_f,
	      cw_nxo_class_delete_t *a_delete_f)
{
    cw_nxoe_class_t *class_;

    class_ = (cw_nxoe_class_t *) nxa_malloc(sizeof(cw_nxoe_class_t));

    nxoe_l_new(&class_->nxoe, NXOT_CLASS, false);
    nxo_null_new(&class_->name);
    nxo_null_new(&class_->super);
    nxo_null_new(&class_->methods);
    nxo_null_new(&class_->data);
    class_->opaque = a_opaque;
    class_->ref_iter_f = a_ref_iter_f;
    class_->delete_f = a_delete_f;

    nxo_no_new(a_nxo);
    a_nxo->o.nxoe = (cw_nxoe_t *) class_;
    nxo_p_type_set(a_nxo, NXOT_CLASS);

    nxa_l_gc_register((cw_nxoe_t *) class_);
}

cw_nxo_t *
nxo_class_name_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_class_t *class_;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CLASS);

    class_ = (cw_nxoe_class_t *) a_nxo->o.nxoe;

    cw_check_ptr(class_);
    cw_dassert(class_->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(class_->nxoe.type == NXOT_CLASS);

    retval = &class_->name;

    return retval;
}

cw_nxo_t *
nxo_class_super_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_class_t *class_;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CLASS);

    class_ = (cw_nxoe_class_t *) a_nxo->o.nxoe;

    cw_check_ptr(class_);
    cw_dassert(class_->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(class_->nxoe.type == NXOT_CLASS);

    retval = &class_->super;

    return retval;
}

cw_nxo_t *
nxo_class_methods_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_class_t *class_;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CLASS);

    class_ = (cw_nxoe_class_t *) a_nxo->o.nxoe;

    cw_check_ptr(class_);
    cw_dassert(class_->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(class_->nxoe.type == NXOT_CLASS);

    retval = &class_->methods;

    return retval;
}

cw_nxo_t *
nxo_class_data_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_class_t *class_;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CLASS);

    class_ = (cw_nxoe_class_t *) a_nxo->o.nxoe;

    cw_check_ptr(class_);
    cw_dassert(class_->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(class_->nxoe.type == NXOT_CLASS);

    retval = &class_->data;

    return retval;
}

void *
nxo_class_opaque_get(const cw_nxo_t *a_nxo)
{
    void *retval;
    cw_nxoe_class_t *class_;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CLASS);

    class_ = (cw_nxoe_class_t *) a_nxo->o.nxoe;

    cw_check_ptr(class_);
    cw_dassert(class_->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(class_->nxoe.type == NXOT_CLASS);

    retval = class_->opaque;

    return retval;
}

void
nxo_class_opaque_set(cw_nxo_t *a_nxo, void *a_opaque)
{
    cw_nxoe_class_t *class_;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CLASS);

    class_ = (cw_nxoe_class_t *) a_nxo->o.nxoe;

    cw_check_ptr(class_);
    cw_dassert(class_->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(class_->nxoe.type == NXOT_CLASS);

    class_->opaque = a_opaque;
}
