/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#define CW_NXO_CONDITION_C_

#include "../include/libonyx/libonyx.h"
#include "../include/libonyx/nxa_l.h"
#include "../include/libonyx/nxo_l.h"
#include "../include/libonyx/nxo_condition_l.h"
#include "../include/libonyx/nxo_mutex_l.h"

void
nxo_condition_new(cw_nxo_t *a_nxo)
{
    cw_nxoe_condition_t *condition;

    condition = (cw_nxoe_condition_t *) nxa_malloc(sizeof(cw_nxoe_condition_t));

    nxoe_l_new(&condition->nxoe, NXOT_CONDITION, false);
    cnd_new(&condition->condition);

    nxo_no_new(a_nxo);
    a_nxo->o.nxoe = (cw_nxoe_t *) condition;
    nxo_p_type_set(a_nxo, NXOT_CONDITION);

    nxa_l_gc_register((cw_nxoe_t *) condition);
}

void
nxo_condition_signal(cw_nxo_t *a_nxo)
{
    cw_nxoe_condition_t *condition;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CONDITION);

    condition = (cw_nxoe_condition_t *) a_nxo->o.nxoe;

    cw_check_ptr(condition);
    cw_dassert(condition->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(condition->nxoe.type == NXOT_CONDITION);

    cnd_signal(&condition->condition);
}

void
nxo_condition_broadcast(cw_nxo_t *a_nxo)
{
    cw_nxoe_condition_t *condition;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CONDITION);

    condition = (cw_nxoe_condition_t *) a_nxo->o.nxoe;

    cw_check_ptr(condition);
    cw_dassert(condition->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(condition->nxoe.type == NXOT_CONDITION);

    cnd_broadcast(&condition->condition);
}

void
nxo_condition_wait(cw_nxo_t *a_nxo, cw_nxo_t *a_mutex)
{
    cw_nxoe_condition_t *condition;
    cw_nxoe_mutex_t *mutex;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CONDITION);

    condition = (cw_nxoe_condition_t *) a_nxo->o.nxoe;

    cw_check_ptr(condition);
    cw_dassert(condition->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(condition->nxoe.type == NXOT_CONDITION);

    cw_check_ptr(a_mutex);
    cw_dassert(a_mutex->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_mutex) == NXOT_MUTEX);

    mutex = (cw_nxoe_mutex_t *) a_mutex->o.nxoe;

    cw_check_ptr(mutex);
    cw_dassert(mutex->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(mutex->nxoe.type == NXOT_MUTEX);

    cnd_wait(&condition->condition, &mutex->lock);
}

bool
nxo_condition_timedwait(cw_nxo_t *a_nxo, cw_nxo_t *a_mutex,
			const struct timespec *a_timeout)
{
    bool retval;
    cw_nxoe_condition_t *condition;
    cw_nxoe_mutex_t *mutex;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_CONDITION);

    condition = (cw_nxoe_condition_t *) a_nxo->o.nxoe;

    cw_check_ptr(condition);
    cw_dassert(condition->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(condition->nxoe.type == NXOT_CONDITION);

    cw_check_ptr(a_mutex);
    cw_dassert(a_mutex->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_mutex) == NXOT_MUTEX);

    mutex = (cw_nxoe_mutex_t *) a_mutex->o.nxoe;

    cw_check_ptr(mutex);
    cw_dassert(mutex->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(mutex->nxoe.type == NXOT_MUTEX);

    retval = cnd_timedwait(&condition->condition, &mutex->lock, a_timeout);

    return retval;
}
