/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#define CW_NXO_HANDLE_C_

#include "../include/libonyx/libonyx.h"
#include "../include/libonyx/nxa_l.h"
#include "../include/libonyx/nxo_l.h"
#include "../include/libonyx/nxo_handle_l.h"

void
nxo_handle_new(cw_nxo_t *a_nxo, void *a_opaque, cw_nxo_handle_eval_t *a_eval_f,
	       cw_nxo_handle_ref_iter_t *a_ref_iter_f,
	       cw_nxo_handle_delete_t *a_delete_f)
{
    cw_nxoe_handle_t *handle;

    handle = (cw_nxoe_handle_t *) nxa_malloc(sizeof(cw_nxoe_handle_t));

    nxoe_l_new(&handle->nxoe, NXOT_HANDLE, false);
    nxo_null_new(&handle->tag);
    handle->opaque = a_opaque;
    handle->eval_f = a_eval_f;
    handle->ref_iter_f = a_ref_iter_f;
    handle->delete_f = a_delete_f;

    nxo_no_new(a_nxo);
    a_nxo->o.nxoe = (cw_nxoe_t *) handle;
    nxo_p_type_set(a_nxo, NXOT_HANDLE);

    nxa_l_gc_register((cw_nxoe_t *) handle);
}

cw_nxo_t *
nxo_handle_tag_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_handle_t *handle;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_HANDLE);

    handle = (cw_nxoe_handle_t *) a_nxo->o.nxoe;

    cw_check_ptr(handle);
    cw_dassert(handle->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(handle->nxoe.type == NXOT_HANDLE);

    retval = &handle->tag;

    return retval;
}

void *
nxo_handle_opaque_get(const cw_nxo_t *a_nxo)
{
    void *retval;
    cw_nxoe_handle_t *handle;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_HANDLE);

    handle = (cw_nxoe_handle_t *) a_nxo->o.nxoe;

    cw_check_ptr(handle);
    cw_dassert(handle->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(handle->nxoe.type == NXOT_HANDLE);

    retval = handle->opaque;

    return retval;
}

void
nxo_handle_opaque_set(cw_nxo_t *a_nxo, void *a_opaque)
{
    cw_nxoe_handle_t *handle;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_HANDLE);

    handle = (cw_nxoe_handle_t *) a_nxo->o.nxoe;

    cw_check_ptr(handle);
    cw_dassert(handle->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(handle->nxoe.type == NXOT_HANDLE);

    handle->opaque = a_opaque;
}

void
nxo_handle_eval(cw_nxo_t *a_nxo, cw_nxo_t *a_thread)
{
    cw_nxoe_handle_t *handle;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_HANDLE);

    handle = (cw_nxoe_handle_t *) a_nxo->o.nxoe;

    cw_check_ptr(handle);
    cw_dassert(handle->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(handle->nxoe.type == NXOT_HANDLE);

    if (handle->eval_f != NULL)
    {
	handle->eval_f(handle->opaque, a_thread);
    }
    else
    {
	cw_nxo_t *ostack, *nxo;

	/* This handle can't be executed, so push it onto ostack just like a
	 * normal literal object would be. */
	ostack = nxo_thread_ostack_get(a_thread);
	nxo = nxo_stack_push(ostack);
	nxo_dup(nxo, a_nxo);
    }
}
