/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#define CW_NXO_INSTANCE_C_

#include "../include/libonyx/libonyx.h"
#include "../include/libonyx/nxa_l.h"
#include "../include/libonyx/nxo_l.h"
#include "../include/libonyx/nxo_instance_l.h"

void
nxo_instance_new(cw_nxo_t *a_nxo, void *a_opaque,
	      cw_nxo_instance_ref_iter_t *a_ref_iter_f,
	      cw_nxo_instance_delete_t *a_delete_f)
{
    cw_nxoe_instance_t *instance;

    instance = (cw_nxoe_instance_t *) nxa_malloc(sizeof(cw_nxoe_instance_t));

    nxoe_l_new(&instance->nxoe, NXOT_INSTANCE, false);
    nxo_null_new(&instance->isa);
    nxo_null_new(&instance->data);
    instance->opaque = a_opaque;
    instance->ref_iter_f = a_ref_iter_f;
    instance->delete_f = a_delete_f;

    nxo_no_new(a_nxo);
    a_nxo->o.nxoe = (cw_nxoe_t *) instance;
    nxo_p_type_set(a_nxo, NXOT_INSTANCE);

    nxa_l_gc_register((cw_nxoe_t *) instance);
}

cw_nxo_t *
nxo_instance_isa_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_instance_t *instance;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_INSTANCE);

    instance = (cw_nxoe_instance_t *) a_nxo->o.nxoe;

    cw_check_ptr(instance);
    cw_dassert(instance->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(instance->nxoe.type == NXOT_INSTANCE);

    retval = &instance->isa;

    return retval;
}

cw_nxo_t *
nxo_instance_data_get(const cw_nxo_t *a_nxo)
{
    cw_nxo_t *retval;
    cw_nxoe_instance_t *instance;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_INSTANCE);

    instance = (cw_nxoe_instance_t *) a_nxo->o.nxoe;

    cw_check_ptr(instance);
    cw_dassert(instance->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(instance->nxoe.type == NXOT_INSTANCE);

    retval = &instance->data;

    return retval;
}

void *
nxo_instance_opaque_get(const cw_nxo_t *a_nxo)
{
    void *retval;
    cw_nxoe_instance_t *instance;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_INSTANCE);

    instance = (cw_nxoe_instance_t *) a_nxo->o.nxoe;

    cw_check_ptr(instance);
    cw_dassert(instance->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(instance->nxoe.type == NXOT_INSTANCE);

    retval = instance->opaque;

    return retval;
}

void
nxo_instance_opaque_set(cw_nxo_t *a_nxo, void *a_opaque)
{
    cw_nxoe_instance_t *instance;

    cw_check_ptr(a_nxo);
    cw_dassert(a_nxo->magic == CW_NXO_MAGIC);
    cw_assert(nxo_type_get(a_nxo) == NXOT_INSTANCE);

    instance = (cw_nxoe_instance_t *) a_nxo->o.nxoe;

    cw_check_ptr(instance);
    cw_dassert(instance->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(instance->nxoe.type == NXOT_INSTANCE);

    instance->opaque = a_opaque;
}
