/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#include "../include/libonyx/libonyx.h"

#include <sys/time.h>
#include <errno.h>

void
tsd_new(cw_tsd_t *a_tsd, void (*a_func)(void *))
{
#ifdef CW_PTHREADS
    int error;
#endif

    cw_check_ptr(a_tsd);

#ifdef CW_PTH
    if (pth_key_create(&a_tsd->key, a_func) == false)
    {
	fprintf(stderr, "%s:%u:%s(): Error in pth_key_create(): %s\n",
		__FILE__, __LINE__, __func__, strerror(errno));
	abort();
    }
#endif
#ifdef CW_PTHREADS
    error = pthread_key_create(&a_tsd->key, a_func);
    if (error)
    {
	fprintf(stderr, "%s:%u:%s(): Error in pthread_key_create(): %s\n",
		__FILE__, __LINE__, __func__, strerror(error));
	abort();
    }
#endif
}

void
tsd_delete(cw_tsd_t *a_tsd)
{
#ifdef CW_PTHREADS
    int error;
#endif

    cw_check_ptr(a_tsd);

#ifdef CW_PTH
    if (pth_key_delete(a_tsd->key) == false)
    {
	fprintf(stderr, "%s:%u:%s(): Error in pth_key_delete(): %s\n",
		__FILE__, __LINE__, __func__, strerror(errno));
	abort();
    }
#endif
#ifdef CW_PTHREADS
    error = pthread_key_delete(a_tsd->key);
    if (error)
    {
	fprintf(stderr, "%s:%u:%s(): Error in pthread_key_delete(): %s\n",
		__FILE__, __LINE__, __func__, strerror(error));
	abort();
    }
#endif
}

void *
tsd_get(cw_tsd_t *a_tsd)
{
    void *retval;

    cw_check_ptr(a_tsd);

#ifdef CW_PTH
    retval = pth_key_getdata(a_tsd->key);
#endif
#ifdef CW_PTHREADS
    retval = pthread_getspecific(a_tsd->key);
#endif

    return retval;
}

void
tsd_set(cw_tsd_t *a_tsd, void *a_val)
{
#ifdef CW_PTHREADS
    int error;
#endif

    cw_check_ptr(a_tsd);

#ifdef CW_PTH
    if (pth_key_setdata(a_tsd->key, a_val) == false)
    {
	fprintf(stderr, "%s:%u:%s(): Error in pth_key_setdata(): %s\n",
		__FILE__, __LINE__, __func__, strerror(errno));
	abort();
    }
#endif
#ifdef CW_PTHREADS
    error = pthread_setspecific(a_tsd->key, a_val);
    if (error)
    {
	fprintf(stderr, "%s:%u:%s(): Error in pthread_setspecific(): %s\n",
		__FILE__, __LINE__, __func__, strerror(error));
	abort();
    }
#endif
}
