/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 * cnd test.
 *
 ******************************************************************************/

#include "../include/libonyx/libonyx.h"

#define CW_TEST_NUM_THREADS 10

struct cw_foo_s
{
    uint32_t *num_waiting;
    cw_cnd_t *cond;
    cw_mtx_t *mutex;
};

void *
thread_entry_func(void *a_arg)
{
    struct cw_foo_s *foo_var = (struct cw_foo_s *) a_arg;

    mtx_lock(foo_var->mutex);
    (*foo_var->num_waiting)++;

    cnd_wait(foo_var->cond, foo_var->mutex);

    fprintf(stderr, "%s(): After cnd_wait() call\n", __func__);

    mtx_unlock(foo_var->mutex);

    return NULL;
}

int
main()
{
    cw_cnd_t cond;
    cw_mtx_t mutex;
    cw_thd_t *threads[CW_TEST_NUM_THREADS], *thread;
    struct cw_foo_s foo_var;
    struct timespec timeout;
    uint32_t i;
    uint32_t num_waiting;

    libonyx_init(0, NULL, NULL);
    fprintf(stderr, "Test begin\n");

    mtx_new(&mutex);

    cnd_new(&cond);
    /* These should do nothing. */
    cnd_signal(&cond);
    cnd_broadcast(&cond);

    /* Set timeout for 1 second. */
    timeout.tv_sec = 1;
    timeout.tv_nsec = 0;

    mtx_lock(&mutex);
    cw_assert(cnd_timedwait(&cond, &mutex, &timeout));
    mtx_unlock(&mutex);

    /* Create argument for thd_new(). */
    num_waiting = 0;
    foo_var.num_waiting = &num_waiting;
    foo_var.cond = &cond;
    foo_var.mutex = &mutex;

    /* Test cnd_signal. */
    thread = thd_new(thread_entry_func, (void *) &foo_var, true);

    /* Bad programming practice, but it works for this test. */
    mtx_lock(&mutex);
    while (num_waiting < 1)
    {
	mtx_unlock(&mutex);
	thd_yield();
	mtx_lock(&mutex);
    }

    cnd_signal(&cond);
    mtx_unlock(&mutex);
    thd_join(thread);

    /* Test cnd_broadcast. */
    num_waiting = 0;
    for (i = 0; i < CW_TEST_NUM_THREADS; i++)
    {
	threads[i] = thd_new(thread_entry_func, (void *) &foo_var, true);
    }

    /* Bad programming practice, but it works for this test. */
    mtx_lock(&mutex);
    while (num_waiting < CW_TEST_NUM_THREADS)
    {
	mtx_unlock(&mutex);
	thd_yield();
	mtx_lock(&mutex);
    }

    cnd_broadcast(&cond);
    mtx_unlock(&mutex);

    for (i = 0; i < CW_TEST_NUM_THREADS; i++)
    {
	thd_join(threads[i]);
    }

    cnd_delete(&cond);
    mtx_delete(&mutex);

    fprintf(stderr, "Test end\n");
    libonyx_shutdown();
    return 0;
}
