/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#include "../include/libonyx/libonyx.h"

/* Number of ring entries (<= 26, preferrably). */
#define NENTRIES 8

typedef struct qstack_s qstack_t;

struct qstack_s {
    qs_elm(qstack_t) link;
    char id;
};

int
main()
{
    uint32_t i;
    qstack_t entries[NENTRIES], *t;
    qs_head(qstack_t) head;

    libonyx_init(0, NULL, NULL);
    fprintf(stderr, "Test begin\n");

    /* Initialize entries. */
    fprintf(stderr, "qs_new(), qs_elm_new()\n");
    qs_new(&head);
    for (i = 0; i < NENTRIES; i++)
    {
	qs_elm_new(&entries[i], link);
	entries[i].id = 'a' + i;
    }
    fprintf(stderr, "  qs_foreach():");
    qs_foreach(t, &head, link)
    {
	fprintf(stderr, " %c", t->id);
    }
    fprintf(stderr, "\n");
    if (qs_top(&head) != NULL)
    {
	fprintf(stderr, "  qs_top(): %c\n", qs_top(&head)->id);
	if (qs_down(qs_top(&head), link) != NULL)
	{
	    fprintf(stderr, "  qs_down(qs_top()): %c\n",
		    qs_down(qs_top(&head), link)->id);
	}
	else
	{
	    fprintf(stderr, "  qs_down(qs_top()): NULL\n");
	}
    }
    else
    {
	fprintf(stderr, "  qs_top(): NULL\n");
    }

    fprintf(stderr, "qs_push()\n");
    for (i = 0; i < NENTRIES; i++)
    {
	qs_push(&head, &entries[i], link);
	fprintf(stderr, "  -->Iteration %u\n", i);
	fprintf(stderr, "  qs_foreach():");
	qs_foreach(t, &head, link)
	{
	    fprintf(stderr, " %c", t->id);
	}
	fprintf(stderr, "\n");
	if (qs_top(&head) != NULL)
	{
	    fprintf(stderr, "  qs_top(): %c\n", qs_top(&head)->id);
	    if (qs_down(qs_top(&head), link) != NULL)
	    {
		fprintf(stderr, "  qs_down(qs_top()): %c\n",
			qs_down(qs_top(&head), link)->id);
	    }
	    else
	    {
		fprintf(stderr, "  qs_down(qs_top()): NULL\n");
	    }
	}
	else
	{
	    fprintf(stderr, "  qs_top(): NULL\n");
	}
    }

    fprintf(stderr, "qs_pop()\n");
    for (i = 0; i < NENTRIES; i++)
    {
	qs_pop(&head, link);
	fprintf(stderr, "  -->Iteration %u\n", i);
	fprintf(stderr, "  qs_foreach():");
	qs_foreach(t, &head, link)
	{
	    fprintf(stderr, " %c", t->id);
	}
	fprintf(stderr, "\n");
	if (qs_top(&head) != NULL)
	{
	    fprintf(stderr, "  qs_top(): %c\n", qs_top(&head)->id);
	    if (qs_down(qs_top(&head), link) != NULL)
	    {
		fprintf(stderr, "  qs_down(qs_top()): %c\n",
			qs_down(qs_top(&head), link)->id);
	    }
	    else
	    {
		fprintf(stderr, "  qs_down(qs_top()): NULL\n");
	    }
	}
	else
	{
	    fprintf(stderr, "  qs_top(): NULL\n");
	}
    }

    fprintf(stderr, "qs_under_push()\n");
    qs_push(&head, &entries[0], link);
    for (i = 1; i < NENTRIES; i++)
    {
	qs_under_push(&entries[0], &entries[i], link);
	fprintf(stderr, "  -->Iteration %u\n", i);
	fprintf(stderr, "  qs_foreach():");
	qs_foreach(t, &head, link)
	{
	    fprintf(stderr, " %c", t->id);
	}
	fprintf(stderr, "\n");
	if (qs_top(&head) != NULL)
	{
	    fprintf(stderr, "  qs_top(): %c\n", qs_top(&head)->id);
	    if (qs_down(qs_top(&head), link) != NULL)
	    {
		fprintf(stderr, "  qs_down(qs_top()): %c\n",
			qs_down(qs_top(&head), link)->id);
	    }
	    else
	    {
		fprintf(stderr, "  qs_down(qs_top()): NULL\n");
	    }
	}
	else
	{
	    fprintf(stderr, "  qs_top(): NULL\n");
	}
    }

    fprintf(stderr, "qs_pop()\n");
    for (i = 0; i < NENTRIES; i++)
    {
	qs_pop(&head, link);
	fprintf(stderr, "  -->Iteration %u\n", i);
	fprintf(stderr, "  qs_foreach():");
	qs_foreach(t, &head, link)
	{
	    fprintf(stderr, " %c", t->id);
	}
	fprintf(stderr, "\n");
	if (qs_top(&head) != NULL)
	{
	    fprintf(stderr, "  qs_top(): %c\n", qs_top(&head)->id);
	    if (qs_down(qs_top(&head), link) != NULL)
	    {
		fprintf(stderr, "  qs_down(qs_top()): %c\n",
			qs_down(qs_top(&head), link)->id);
	    }
	    else
	    {
		fprintf(stderr, "  qs_down(qs_top()): NULL\n");
	    }
	}
	else
	{
	    fprintf(stderr, "  qs_top(): NULL\n");
	}
    }

    fprintf(stderr, "Test end\n");
    libonyx_shutdown();
    return 0;
}
