/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2004 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.1
 *
 ******************************************************************************/

#include "../include/libonyx/libonyx.h"

#define NSUSPENSIBLE 10
#define NSUSPENDERS 10
#define NTHREADS (NSUSPENSIBLE + NSUSPENDERS)
#define NITERATIONS 100

uint32_t count = 0;

void *
thread_entry_func(void *a_arg)
{
    uint32_t i;

    for (i = 0; i < NITERATIONS; i++)
    {
	thd_single_enter();
	count++;
	thd_single_leave();
    }

    return NULL;
}

int
main()
{
    cw_thd_t *thds[NTHREADS];
    uint32_t i;

    libonyx_init(0, NULL, NULL);
    fprintf(stderr, "Test begin\n");

    /* Create threads that can be suspended. */
    for (i = 0; i < NSUSPENSIBLE; i++)
    {
	thds[i] = thd_new(thread_entry_func, NULL, true);
    }
    /* Create threads that cannot be suspended. */
    for (; i < NTHREADS; i++)
    {
	thds[i] = thd_new(thread_entry_func, NULL, false);
    }

    /* The initial thread is suspensible. */
    thread_entry_func(NULL);

    for (i = 0; i < NTHREADS; i++)
    {
	thd_join(thds[i]);
    }

    cw_assert(count == (NTHREADS + 1) * NITERATIONS);

    fprintf(stderr, "Test end\n");
    libonyx_shutdown();
    return 0;
}
